"""Home to some data readers."""

from abc import ABC, abstractmethod
import pandas as pd
from typing import List, Dict

from .flodym_arrays import Parameter
from .mfa_definition import DimensionDefinition, ParameterDefinition
from .dimensions import DimensionSet, Dimension


class DataReader:
    """Template for creating a data reader, showing required methods and data formats needed for
    use in the MFASystem model.
    """

    def read_dimensions(self, dimension_definitions: List[DimensionDefinition]) -> DimensionSet:
        """Method to read data for multiple dimensions, by looping over `read_dimension`."""
        dimensions = [self.read_dimension(definition) for definition in dimension_definitions]
        return DimensionSet(dim_list=dimensions)

    @abstractmethod
    def read_dimension(self, dimension_definition: DimensionDefinition) -> Dimension:
        """Required method to read data for a single dimension,
        corresponding to the dimension definition."""
        pass

    @abstractmethod
    def read_parameter_values(self, parameter_name: str, dims: DimensionSet) -> Parameter:
        """Required method to read data for a particular parameter."""
        pass

    def read_parameters(
        self, parameter_definitions: List[ParameterDefinition], dims: DimensionSet
    ) -> Dict[str, Parameter]:
        """Method to read data for a list of parameters, by looping over `read_parameter_values`."""
        parameters = {}
        for parameter_definition in parameter_definitions:
            dim_subset = dims.get_subset(parameter_definition.dim_letters)
            parameters[parameter_definition.name] = self.read_parameter_values(
                parameter_name=parameter_definition.name,
                dims=dim_subset,
            )
        return parameters


class DimensionReader(ABC):
    """Template for creating a dimension reader, showing required methods and data formats needed"""

    read_dimensions = DataReader.read_dimensions

    @abstractmethod
    def read_dimension(self, dimension_definition: DimensionDefinition) -> Dimension:
        pass


class CSVDimensionReader(DimensionReader):
    """Read dimensions from a CSV file. Expects a single row or single columns csv file with no header containing the dimension items.

    Args:
        dimension_files (dict): {dimension_name: file_path, ...}
        read_csv_kwargs: Additional keyword arguments passed to pandas.read_csv.
            The default is {"header": None}. Not encouraged to use, since it may not lead to the
            intended DataFrame format. Sticking to recommended csv file format is preferred.
    """

    def __init__(
        self,
        dimension_files,
        **read_csv_kwargs,
    ):
        self.dimension_files = dimension_files
        self.read_csv_kwargs = read_csv_kwargs

    def read_dimension(self, definition: DimensionDefinition):
        path = self.dimension_files[definition.name]
        if "header" not in self.read_csv_kwargs:
            self.read_csv_kwargs["header"] = None
        df = pd.read_csv(path, **self.read_csv_kwargs)
        return Dimension.from_df(df, definition)


class ExcelDimensionReader(DimensionReader):
    """Read dimensions from Excel file(s). Expects a single row or single columns excel sheet with no header containing the dimension items.

    Args:
        dimension_files (dict): {dimension_name: file_path, ...}
        dimension_sheets (dict): {dimension_name: sheet_name, ...}
        ead_excel_kwargs: Additional keyword arguments passed to pandas.read_excel.
            The default is {"header": None}. Not encouraged to use, since it may not lead to the
            intended DataFrame format. Sticking to recommended excel file format is preferred.
    """

    def __init__(
        self,
        dimension_files: dict,
        dimension_sheets: dict = None,
        **read_excel_kwargs,
    ):
        self.dimension_files = dimension_files
        self.dimension_sheets = dimension_sheets
        self.read_excel_kwargs = read_excel_kwargs

    def read_dimension(self, definition: DimensionDefinition):
        path = self.dimension_files[definition.name]
        # load data from excel
        if self.dimension_sheets is None:
            sheet_name = None
        else:
            sheet_name = self.dimension_sheets[definition.name]
        # default for header is None
        if "header" not in self.read_excel_kwargs:
            self.read_excel_kwargs["header"] = None
        df = pd.read_excel(path, sheet_name=sheet_name, **self.read_excel_kwargs)
        return Dimension.from_df(df, definition)


class ParameterReader(ABC):
    """Template for creating a parameter reader, showing required methods and data formats needed"""

    @abstractmethod
    def read_parameter_values(self, parameter_name: str, dims: DimensionSet) -> Parameter:
        pass

    read_parameters = DataReader.read_parameters


class CSVParameterReader(ParameterReader):
    """Reads a csv file to a pandas data frame and calls :py:meth:`flodym.Parameter.from_df`.
    Expects comma separation and no header, apart from optional column names.
    For further detail on expected format, see :py:meth:`flodym.FlodymArray.from_df`.

    Args:
        parameter_files (dict): Mapping of parameter names to file paths.
            Format: {parameter_name: file_path, ...}
        allow_missing_values (bool, optional): Whether to allow missing values in the DataFrame.
            This includes both missing rows, and NaN values in the value column.
            Defaults to False.
        allow_extra_values (bool, optional): Whether to allow extra rows in the DataFrame,
            i.e. tows with index items not present in the FlodymArray dimension items.
            Defaults to False.
        read_csv_kwargs: Additional keyword arguments passed to pandas.read_csv.
            Not encouraged to use, since it may not lead to the intended DataFrame format.
            Sticking to recommended csv file format is preferred.
    """

    def __init__(
        self,
        parameter_files: dict = None,
        allow_missing_values: bool = False,
        allow_extra_values: bool = False,
        **read_csv_kwargs,
    ):
        self.parameter_filenames = parameter_files  # {parameter_name: file_path, ...}
        self.allow_missing_values = allow_missing_values
        self.allow_extra_values = allow_extra_values
        self.read_csv_kwargs = read_csv_kwargs

    def read_parameter_values(self, parameter_name: str, dims):
        if self.parameter_filenames is None:
            raise ValueError("No parameter files specified.")
        datasets_path = self.parameter_filenames[parameter_name]
        data = pd.read_csv(datasets_path, **self.read_csv_kwargs)
        return Parameter.from_df(
            dims=dims,
            name=parameter_name,
            df=data,
            allow_missing_values=self.allow_missing_values,
            allow_extra_values=self.allow_extra_values,
        )


class ExcelParameterReader(ParameterReader):
    """Reads an excel file to a pandas data frame and calls :py:meth:`flodym.Parameter.from_df`.
    Expects contiguous data starting in the upper left cell A1.
    For further detail on expected format, see :py:meth:`flodym.FlodymArray.from_df`.

    Args:
        parameter_files (dict): Mapping of parameter names to file paths.
            Can be the same file for multiple parameters if the sheets are different.
            Format: {parameter_name: file_path, ...}
        parameter_sheets (dict): Mapping of parameter names to sheet names in the excel file.
            If None, the first sheet is used.
            Format: {parameter_name: sheet_name, ...}
            Defaults to None.
        allow_missing_values (bool, optional): Whether to allow missing values in the DataFrame.
            This includes both missing rows, and NaN values in the value column.
            Defaults to False.
        allow_extra_values (bool, optional): Whether to allow extra rows in the DataFrame,
            i.e. tows with index items not present in the FlodymArray dimension items.
            Defaults to False.
        read_excel_kwargs: Additional keyword arguments passed to pandas.read_excel.
            Not encouraged to use, since it may not lead to the intended DataFrame format.
            Sticking to recommended excel file format is preferred
    """

    def __init__(
        self,
        parameter_files: dict,
        parameter_sheets: dict = None,
        allow_missing_values: bool = False,
        allow_extra_values: bool = False,
        **read_excel_kwargs,
    ):
        self.parameter_files = parameter_files
        self.parameter_sheets = parameter_sheets
        self.allow_missing_values = allow_missing_values
        self.allow_extra_values = allow_extra_values
        self.read_excel_kwargs = read_excel_kwargs

    def read_parameter_values(self, parameter_name: str, dims):
        datasets_path = self.parameter_files[parameter_name]
        if self.parameter_sheets is None:
            sheet_name = None
        else:
            sheet_name = self.parameter_sheets[parameter_name]
        data = pd.read_excel(datasets_path, sheet_name=sheet_name, **self.read_excel_kwargs)
        return Parameter.from_df(
            dims=dims,
            name=parameter_name,
            df=data,
            allow_missing_values=self.allow_missing_values,
            allow_extra_values=self.allow_extra_values,
        )


class CompoundDataReader(DataReader):
    """Combines a DimensionReader and a ParameterReader to create a DataReader,
    reading both dimensions and parameters.
    """

    def __init__(
        self,
        dimension_reader: DimensionReader,
        parameter_reader: ParameterReader,
    ):
        self.dimension_reader = dimension_reader
        self.parameter_reader = parameter_reader

    def read_dimension(self, dimension_definition: DimensionDefinition) -> Dimension:
        return self.dimension_reader.read_dimension(dimension_definition)

    def read_parameter_values(self, parameter_name: str, dims: DimensionSet) -> Parameter:
        return self.parameter_reader.read_parameter_values(parameter_name, dims)



from .processes import Process


def process_names_with_arrow(from_process: Process, to_process: Process) -> str:
    """Return the name of a flow as a string containing
    the process it starts from and the process it ends in, separated by an arrow.
    """
    return f"{from_process.name} => {to_process.name}"


def process_names_no_spaces(from_process: Process, to_process: Process) -> str:
    """Return the name of a flow as a string containing
    the process it starts from and the process it ends in, separated by '_to_', without spaces.
    """
    return f"{from_process.name.replace(' ', '_')}_to_{to_process.name.replace(' ', '_')}"


def process_ids(from_process: Process, to_process: Process) -> str:
    """Return the name of a flow as a string containing the id of the process it starts from and the id of the process it ends in, e.g. F1_2."""
    return f"F{from_process.id}_{to_process.id}"



import numpy as np
import pytest

from flodym.data_reader import (
    CSVDimensionReader,
    ExcelDimensionReader,
    CSVParameterReader,
    ExcelParameterReader,
)
from flodym.mfa_definition import DimensionDefinition, ParameterDefinition, MFADefinition
from flodym.mfa_system import MFASystem


csv_dimension_files = {
    "animals": "tests/tests_data/dimension_animals_horizontal.csv",
    "time": "tests/tests_data/dimension_time_vertical.csv",
    "region": "tests/tests_data/dimension_region_single.csv",
}
excel_dimension_file = "tests/tests_data/dimensions.xlsx"
excel_dimension_files = {a: excel_dimension_file for a in csv_dimension_files.keys()}
dimension_sheet_names = {a: a for a in csv_dimension_files.keys()}
dimension_definitions = [
    DimensionDefinition(name="animals", letter="a", dtype=str),
    DimensionDefinition(name="time", letter="t", dtype=int),
    DimensionDefinition(name="region", letter="r", dtype=str),
]
dimension_items_expected = {
    "animals": ["dog", "cat"],
    "time": [1990, 2000, 2010],
    "region": ["world"],
}
parameter_definitions = [
    ParameterDefinition(name="p1", dim_letters=["t", "r"]),
    ParameterDefinition(name="p2", dim_letters=["t", "r"]),
    ParameterDefinition(name="p3", dim_letters=["t", "r"]),
    ParameterDefinition(name="p4", dim_letters=["t", "r"]),
    ParameterDefinition(name="p5", dim_letters=["t", "r", "a"]),
    ParameterDefinition(name="p6", dim_letters=["t", "r", "a"]),
]
csv_parameter_files = {
    a.name: "tests/tests_data/parameter_" + a.name + ".csv" for a in parameter_definitions
}
excel_parameter_file = "tests/tests_data/parameters.xlsx"
excel_parameter_files = {a.name: excel_parameter_file for a in parameter_definitions}
parameter_sheet_names = {a.name: a.name for a in parameter_definitions}


wrong_parameter_definitions = [
    # missing row
    ParameterDefinition(name="e1", dim_letters=["t", "a"]),
    # wrong dim item
    ParameterDefinition(name="e2", dim_letters=["t"]),
    # additional dim item
    ParameterDefinition(name="e3", dim_letters=["t"]),
    # empty cell
    ParameterDefinition(name="e4", dim_letters=["t"]),
    # missing dim
    ParameterDefinition(name="e5", dim_letters=["t", "a"]),
]

wrong_csv_parameter_files = {
    a.name: "tests/tests_data/parameter_" + a.name + ".csv" for a in wrong_parameter_definitions
}
wrong_excel_parameter_file = "tests/tests_data/parameters.xlsx"
wrong_excel_parameter_files = {
    a.name: wrong_excel_parameter_file for a in wrong_parameter_definitions
}
wrong_sheet_names = {a.name: a.name for a in wrong_parameter_definitions}


def test_dimensions():

    csv_reader = CSVDimensionReader(csv_dimension_files)
    excel_reader = ExcelDimensionReader(excel_dimension_files, dimension_sheet_names)

    for reader in [csv_reader, excel_reader]:
        dimension_set = reader.read_dimensions(dimension_definitions)

        for dim in dimension_set:
            assert dim.items == dimension_items_expected[dim.name]


def test_valid_parameter_reader():

    dims = CSVDimensionReader(csv_dimension_files).read_dimensions(dimension_definitions)

    csv_reader = CSVParameterReader(csv_parameter_files)
    excel_reader = ExcelParameterReader(excel_parameter_files, parameter_sheet_names)

    for reader in [csv_reader, excel_reader]:
        # for reader in [excel_reader]:
        parameters = reader.read_parameters(parameter_definitions, dims)
        for pn, p in parameters.items():
            if pn in ["p1", "p2", "p3", "p4"]:
                assert np.array_equal(p.values, [[1], [2], [3]])
            elif pn in ["p5", "p6"]:
                assert np.array_equal(p.values, np.array([[[1, 4]], [[2, 5]], [[3, 6]]]))


def test_wrong_parameter_reader():

    wrong_csv_reader = CSVParameterReader(wrong_csv_parameter_files)
    wrong_excel_reader = ExcelParameterReader(wrong_excel_parameter_files, wrong_sheet_names)

    dims = CSVDimensionReader(csv_dimension_files).read_dimensions(dimension_definitions)

    for reader in [wrong_csv_reader, wrong_excel_reader]:
        for prm_def in wrong_parameter_definitions:
            with pytest.raises(ValueError):
                reader.read_parameter_values(prm_def.name, dims.get_subset(prm_def.dim_letters))


def test_allow_incomplete_data():
    incomplete_parameter_definitions = [
        # missing row
        ParameterDefinition(name="e1", dim_letters=["t", "a"]),
        # empty cell
        ParameterDefinition(name="e4", dim_letters=["t"]),
    ]

    incomplete_csv_reader = CSVParameterReader(wrong_csv_parameter_files, allow_missing_values=True)
    incomplete_excel_reader = ExcelParameterReader(
        wrong_excel_parameter_files, wrong_sheet_names, allow_missing_values=True
    )

    dims = CSVDimensionReader(csv_dimension_files).read_dimensions(dimension_definitions)

    for reader in [incomplete_csv_reader, incomplete_excel_reader]:
        parameters = reader.read_parameters(incomplete_parameter_definitions, dims)
        assert np.array_equal(parameters["e1"].values, [[4, 1], [5, 0], [6, 3]])
        assert np.array_equal(parameters["e4"].values, [0, 2, 3])


def test_allow_extra_data():
    extra_parameter_definitions = [
        # wrong dim item
        ParameterDefinition(name="e2", dim_letters=["t"]),
        # additional dim item
        ParameterDefinition(name="e3", dim_letters=["t"]),
    ]

    extra_csv_reader = CSVParameterReader(
        wrong_csv_parameter_files, allow_missing_values=True, allow_extra_values=True
    )
    extra_excel_reader = ExcelParameterReader(
        wrong_excel_parameter_files,
        wrong_sheet_names,
        allow_missing_values=True,
        allow_extra_values=True,
    )

    dims = CSVDimensionReader(csv_dimension_files).read_dimensions(dimension_definitions)

    for reader in [extra_csv_reader, extra_excel_reader]:
        parameters = reader.read_parameters(extra_parameter_definitions, dims)
        assert np.array_equal(parameters["e2"].values, [1, 0, 3])
        assert np.array_equal(parameters["e3"].values, [1, 2, 3])


def test_build_mfa_system():

    definition = MFADefinition(
        dimensions=dimension_definitions,
        processes=[],
        stocks=[],
        flows=[],
        parameters=parameter_definitions,
    )

    class MinimalMFASystem(MFASystem):
        def compute(self):
            pass

    mfa = MinimalMFASystem.from_csv(
        definition,
        csv_dimension_files,
        csv_parameter_files,
    )

    mfa = MinimalMFASystem.from_excel(
        definition,
        excel_dimension_files,
        excel_parameter_files,
        dimension_sheets=dimension_sheet_names,
        parameter_sheets=parameter_sheet_names,
    )


if __name__ == "__main__":
    test_dimensions()
    test_valid_parameter_reader()
    test_wrong_parameter_reader()
    test_allow_incomplete_data()
    test_allow_extra_data()
    test_build_mfa_system()
    print("All tests passed.")



import sys
import logging
import numpy as np
import pandas as pd
from typing import Literal, Optional, TYPE_CHECKING, Iterable
from pydantic import BaseModel as PydanticBaseModel
import itertools

from .dimensions import Dimension

if TYPE_CHECKING:
    from .flodym_arrays import FlodymArray


class FlodymDataFormat(PydanticBaseModel):

    type: Literal["long", "wide"]
    value_column: str = "value"
    columns_dim: Optional[str] = None


class DataFrameToFlodymDataConverter:
    """Converts a panda DataFrame with various possible formats to a numpy array that can be used
    as values of a FlodymArray.

    Usually not called by the user, but from within the FlodymArray from_df and
    set_values_from_df methods.

    In case of errors, turning on debug logging might help to understand the process.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        flodym_array: "FlodymArray",
        allow_missing_values: bool = False,
        allow_extra_values: bool = False,
    ):
        self.df = df.copy()
        self.flodym_array = flodym_array
        self.allow_missing_values = allow_missing_values
        self.allow_extra_values = allow_extra_values
        try:
            self.target_values = self.get_target_values()
        except Exception as e:
            # add error context to all errors to ease debugging
            raise type(e)(f"{self.error_context} {str(e)}").with_traceback(sys.exc_info()[2])

    def get_target_values(self) -> np.ndarray:
        logging.debug(
            f"Start setting values for FlodymArray {self.flodym_array.name} with dimensions {self.flodym_array.dims.names} from dataframe."
        )
        self._reset_non_default_index()
        self._determine_format()
        self._df_to_long_format()
        self._check_missing_dim_columns()
        self._convert_type()
        self._sort_columns()
        values = self._check_data_complete()
        return values

    def _reset_non_default_index(self):
        if isinstance(self.df.index, pd.MultiIndex):
            self.df.reset_index(inplace=True)
        elif self.df.index.name is not None:
            self.df.reset_index(inplace=True)
        elif self.df.index.dtype != np.int64:
            self.df.reset_index(inplace=True)
        elif self.df.index.min() >= 1700 and self.df.index.max() <= 2300:
            self.df.reset_index(inplace=True)

    def _determine_format(self):
        self._get_dim_columns_by_name()
        self._check_if_first_row_are_items()
        self._check_for_dim_columns_by_items()
        self._check_value_columns()

    def _get_dim_columns_by_name(self):
        self.dim_columns = [c for c in self.df.columns if c in self.flodym_array.dims.names]
        logging.debug(f"Recognized index columns by name: {self.dim_columns}")

    def _check_if_first_row_are_items(self):
        """If data without columns names was read, but the first row was assumed to be column names,
        the first row of the data frame might erroneously end up as column names.
        This method checks if that is the case, and if so, prepends a row based on column names.
        """
        column_name = self.df.columns[0]
        col_items = self.df[column_name].unique()
        extended_col_items = [column_name] + col_items.tolist()
        for dim in self.flodym_array.dims:
            if self.same_items(extended_col_items, dim):
                self._add_column_names_as_row(column_name, dim)

    def _add_column_names_as_row(self, column_name: str, dim: Dimension):
        if len(self.dim_columns) > 0:
            raise ValueError(
                f"Ambiguouity detected: column with first item {column_name} could be "
                f"dimension {dim.name} if the first row counts as an item, but columns "
                f"{self.dim_columns} are already recognized as dimensions first row as name."
                f" Please change the item names of the affected dimension, or use a"
                f" different method to read data."
            )
        # prepend a row to df with all the column names
        self.df = pd.concat([pd.DataFrame([self.df.columns], columns=self.df.columns), self.df])
        # rename columns to range from 0 to n, save original name
        column_name = f"column {self.df.columns.get_loc(column_name)}"
        self.df.columns = [f"column {i}" for i in range(len(self.df.columns))]

    def _check_for_dim_columns_by_items(self):
        for cn in self.df.columns:
            if cn in self.dim_columns:
                continue
            found = self._check_if_dim_column_by_items(cn)
            if not found:
                logging.debug(
                    f"Could not find dimension with same items as column {cn}. "
                    "Assuming this is the first value column; Won't look further."
                )
                return

    def _check_if_dim_column_by_items(self, column_name: str) -> bool:
        logging.debug(f"Checking if {column_name} is a dimension by comparing items with dim items")
        col_items = self.df[column_name].unique()
        for dim in self.flodym_array.dims:
            if self.same_items(col_items, dim):
                logging.debug(f"{column_name} is dimension {dim.name}.")
                self.df.rename(columns={column_name: dim.name}, inplace=True)
                self.dim_columns.append(dim.name)
                return True
        return False

    def _check_value_columns(self):
        value_cols = np.setdiff1d(list(self.df.columns), self.dim_columns)
        logging.debug(f"Assumed value columns: {value_cols}")
        value_cols_are_dim_items = self._check_if_value_columns_match_dim_items(value_cols)
        if not value_cols_are_dim_items:
            self._check_if_valid_long_format(value_cols)

    def _check_if_value_columns_match_dim_items(self, value_cols: list[str]) -> bool:
        logging.debug("Trying to match set of value column names with items of dimension.")
        for dim in self.flodym_array.dims:
            if self.same_items(value_cols, dim):
                logging.debug(f"Value columns match dimension items of {dim.name}.")
                self.format = FlodymDataFormat(type="wide", columns_dim=dim.name)
                if dim.dtype is not None:
                    for c in value_cols:
                        self.df.rename(columns={c: dim.dtype(c)}, inplace=True)
                return True
        return False

    def _check_if_valid_long_format(self, value_cols: list[str]):
        logging.debug(
            "Could not find dimension with same item set as value column names. Assuming long format, i.e. one value column."
        )
        if len(value_cols) == 1:
            self.format = FlodymDataFormat(type="long", value_column=value_cols[0])
            logging.debug(f"Value column name is {value_cols[0]}.")
        else:
            raise ValueError(
                f"More than one value columns. Could not find a dimension the items of which match the set of value column names. "
                f"Value columns: {value_cols}. Please check input data for format, typos, data types and missing items."
            )

    def _df_to_long_format(self):
        if self.format.type != "wide":
            return
        logging.debug("Converting wide format to long format.")
        value_cols = self.flodym_array.dims[self.format.columns_dim].items
        self.df = self.df.melt(
            id_vars=[c for c in self.df.columns if c not in value_cols],
            value_vars=value_cols,
            var_name=self.format.columns_dim,
            value_name=self.format.value_column,
        )
        self.dim_columns.append(self.format.columns_dim)
        self.format = FlodymDataFormat(type="long", value_column=self.format.value_column)

    def _check_missing_dim_columns(self):
        missing_dim_columns = np.setdiff1d(list(self.flodym_array.dims.names), self.dim_columns)
        for c in missing_dim_columns:
            if len(self.flodym_array.dims[c].items) == 1:
                self.df[c] = self.flodym_array.dims[c].items[0]
                self.dim_columns.append(c)
            else:
                raise ValueError(
                    f"Dimension {c} from array has more than one item, but is not found in df. Please specify column in dataframe."
                )

    def _convert_type(self):
        for dim in self.flodym_array.dims:
            if dim.dtype is not None:
                self.df[dim.name] = self.df[dim.name].map(dim.dtype)
        self.df[self.format.value_column] = self.df[self.format.value_column].astype(np.float64)

    def _sort_columns(self):
        """Sort the columns of the data frame according to the order of the dimensions in the
        FlodymArray.
        """
        self.df = self.df[list(self.flodym_array.dims.names) + [self.format.value_column]]

    def _check_data_complete(self):
        # check for double entries in the index columns
        indices = self.df[list(self.flodym_array.dims.names)]
        if indices.duplicated().any():
            raise ValueError(
                f"The following index combinations occur more than once in the data: ",
                indices[indices.duplicated()],
            )

        # remove rows with extra values or throw error
        if self.allow_extra_values:
            for dim in self.flodym_array.dims:
                self.df = self.df[self.df[dim.name].isin(dim.items)]
        else:
            for dim in self.flodym_array.dims:
                unique_items = set(self.df[dim.name].unique())
                extra_items = unique_items - set(dim.items)
                if extra_items:
                    raise ValueError(
                        f"Dimension column '{dim.name}' contains items that are not in the dimension: {extra_items}"
                    )

        if self.allow_missing_values:
            # fill na values with zeros
            self.df[self.format.value_column].fillna(0, inplace=True)
        else:
            if len(self.df) != self.flodym_array.size:
                # print warning first, as compiling expected index tuples may take long
                logging.warning(
                    f"{self.error_context} Detected missing values in the data, but "
                    f"allow_missing_values is set to False. Expected {self.flodym_array.size} "
                    f"rows, but only got {len(self.df)}. Computing missing values...."
                )
                expected_index_tuples = set(
                    itertools.product(*[dim.items for dim in self.flodym_array.dims])
                )
                indices = self.df[list(self.flodym_array.dims.names)]
                actual_index_tuples = set(indices.itertuples(index=False, name=None))
                unexpected_items = actual_index_tuples - expected_index_tuples
                raise ValueError(
                    f"Detected missing values in the data, but allow_missing_values is set to False. "
                    f"Missing values for index combinations: {unexpected_items}."
                )
            if any(self.df[self.format.value_column].isna()):
                raise ValueError("Empty cells/NaN values in value column!")

        # for performance and memory reasons, we use numpy operations to convert the data to the numpy array
        # replace non-value items in df with their index in dims.items
        for dim in self.flodym_array.dims:
            self.df[dim.name] = self.df[dim.name].map({item: i for i, item in enumerate(dim.items)})
        # convert df to numpy index array: Take all non-value columns
        fill_indices = self.df.values[:, :-1].T.astype(np.int16)
        fill_values = self.df[self.format.value_column].values
        # this is what ends up in the parameter; initialize with zeros
        values = np.zeros(self.flodym_array.dims.shape())
        values[tuple(fill_indices)] = fill_values
        return values

    @staticmethod
    def same_items(arr: Iterable, dim: Dimension) -> bool:
        if dim.dtype is not None:
            try:
                arr = [dim.dtype(a) for a in arr]
            except ValueError:
                return False
        return len(set(arr).symmetric_difference(set(dim.items))) == 0

    @property
    def error_context(self) -> str:
        return f"While setting values of {self.flodym_array.__class__.__name__} '{self.flodym_array.name}' from DataFrame:"



from flodym.mfa_definition import (
    MFADefinition as MFADefinition,
    FlowDefinition as FlowDefinition,
    DimensionDefinition as DimensionDefinition,
    StockDefinition as StockDefinition,
    ParameterDefinition as ParameterDefinition,
)
from flodym.mfa_system import MFASystem as MFASystem
from flodym.dimensions import DimensionSet as DimensionSet, Dimension as Dimension
from flodym.flodym_arrays import (
    FlodymArray as FlodymArray,
    Parameter as Parameter,
    StockArray as StockArray,
    Flow as Flow,
)
from flodym.processes import Process as Process, make_processes as make_processes
from flodym.stocks import (
    Stock as Stock,
    SimpleFlowDrivenStock as SimpleFlowDrivenStock,
    DynamicStockModel as DynamicStockModel,
    InflowDrivenDSM as InflowDrivenDSM,
    StockDrivenDSM as StockDrivenDSM,
    StockDrivenDSM_NIC as StockDrivenDSM_NIC,
)
from flodym.lifetime_models import (
    LifetimeModel as LifetimeModel,
    FixedLifetime as FixedLifetime,
    NormalLifetime as NormalLifetime,
    FoldedNormalLifetime as FoldedNormalLifetime,
    LogNormalLifetime as LogNormalLifetime,
    WeibullLifetime as WeibullLifetime,
)
from flodym.flow_helper import make_empty_flows as make_empty_flows
from flodym.stock_helper import make_empty_stocks as make_empty_stocks
from flodym.data_reader import (
    DataReader as DataReader,
    DimensionReader as DimensionReader,
    CSVDimensionReader as CSVDimensionReader,
    ExcelDimensionReader as ExcelDimensionReader,
    ParameterReader as ParameterReader,
    CSVParameterReader as CSVParameterReader,
    ExcelParameterReader as ExcelParameterReader,
    CompoundDataReader as CompoundDataReader,
)



from .data_writer import (
    export_mfa_flows_to_csv as export_mfa_flows_to_csv,
    export_mfa_stocks_to_csv as export_mfa_stocks_to_csv,
    export_mfa_to_pickle as export_mfa_to_pickle,
    convert_to_dict as convert_to_dict,
)
from .array_plotter import (
    ArrayPlotter as ArrayPlotter,
    PlotlyArrayPlotter as PlotlyArrayPlotter,
    PyplotArrayPlotter as PyplotArrayPlotter,
)
from .sankey import PlotlySankeyPlotter as PlotlySankeyPlotter



from copy import copy
from pydantic import BaseModel as PydanticBaseModel, Field, AliasChoices, model_validator
from typing import Dict, Iterator, Optional
import numpy as np
import pandas as pd

from .mfa_definition import DimensionDefinition


class Dimension(PydanticBaseModel):
    """One of multiple dimensions over which MFA arrays are defined.

    Defined by a name, a letter for shorter addressing, and a list of items.

    **Example**

    >>> from flodym import Dimension
    >>> regions = Dimension(name='Region', letter='r', items=['Earth', 'Moon', 'Sun'])

    The list of items can be loaded using a :py:class:`flodym.DataReader` object,
    or set directly, for example if a subset of an existing dimension is formed.
    """

    name: str = Field(..., min_length=2)
    """The full name of the dimension"""
    letter: str = Field(
        ..., min_length=1, max_length=1, validation_alias=AliasChoices("letter", "dim_letter")
    )
    """A single index letter for shorter addressing of the dimension"""
    items: list
    """A list of items that are resolved along the dimension, e.g. years or regions"""
    dtype: Optional[type] = None
    """If given, a check is performed that all items have this datatype. Recommended for safety,
    especially in ambiguous cases such as calendar years (int or str)
    """

    @model_validator(mode="after")
    def items_have_datatype(self):
        """If a datatype is specified, check that all items have this datatype."""
        if self.dtype is not None and any([not isinstance(i, self.dtype) for i in self.items]):
            raise ValueError("All items must have the same datatype as specified in dtype.")
        return self

    @classmethod
    def from_df(cls, df: pd.DataFrame, definition: DimensionDefinition) -> "Dimension":
        """Create Dimension object from definition and a pandas DataFrame for the items.

        Args:
            df (pd.DataFrame): A single-column or single-row data frame, which is transformed to the dimension items.
            definition (DimensionDefinition): The definition of the dimension.

        Returns:
            Dimension: The dimension object.
        """
        data = df.to_numpy()
        return cls.from_np(data, definition)

    @classmethod
    def from_np(cls, data: np.ndarray, definition: DimensionDefinition) -> "Dimension":
        """Create Dimension object from definition and a numpy array for the items.
        Performs checks on the array and transforms to list of items.

        Args:
            data (np.ndarray): A numpy array where only one dimension exceeds size 1.
            definition (DimensionDefinition): The definition object

        Returns:
            Dimension: The dimension object.
        """
        if sum([s > 1 for s in data.shape]) > 1:
            raise ValueError(
                f"Dimension data for {definition.name} must have only one row or column."
            )
        data = data.flatten().tolist()
        # delete header for items if present
        if data[0] == definition.name:
            data = data[1:]
        data = [definition.dtype(item) for item in data]
        return cls(
            name=definition.name, letter=definition.letter, items=data, dtype=definition.dtype
        )

    @property
    def len(self) -> int:
        """The number of items in the dimension."""
        return len(self.items)

    def index(self, item) -> int:
        """Return the index of an item in the dimension."""
        return self.items.index(item)

    def is_subset(self, other: "Dimension"):
        """Check if the items of this dimension are a subset of the items of another dimension."""
        return set(self.items).issubset(other.items)

    def is_superset(self, other: "Dimension"):
        """Check if the items of this dimension are a superset of the items of another dimension."""
        return set(self.items).issuperset(other.items)


class DimensionSet(PydanticBaseModel):
    """A set of Dimension objects which MFA arrays are defined over.

    **Example**

        >>> from flodym import Dimension, DimensionSet
        >>> regions = Dimension(name='Region', letter='r', items=['Earth', 'Moon', 'Sun'])
        >>> time = Dimension(name='Time', letter='t', items=[1990, 2000, 2010, 2020, 2030])
        >>> dimensions = DimensionSet([regions, time])

    It is expected that DimensionSet instances are created via the :py:class:`flodym.DataReader`.

        >>> from flodym import DataReader, DimensionDefinition, Dimension
        >>> class MyDataReader(DataReader):
        >>>    def read_dimension(self, dimension_definition: DimensionDefinition) -> Dimension:
        >>>        if dimension_definition.letter == 't':
        >>>            return Dimension(name='Time', letter='t', items=[1990, 2000, 2010, 2020, 2030])
        >>>        elif dimension_definition.letter == 'r':
        >>>            return Dimension(name='Region', letter='r', items=['Earth', 'Moon', 'Sun'])
        >>>        raise ValueError('No data available for desired dimension')
        >>> data_reader = MyDataReader()
        >>> time_definition = DimensionDefinition(name='Time', letter='t', dtype=int)
        >>> region_definition = DimensionDefinition(name='Region', letter='r', dtype=str)
        >>> definitions = [time_definition, region_definition]
        >>> dimensions = data_reader.read_dimensions(dimension_definitions=definitions)

    """

    dim_list: list[Dimension]
    """A list of Dimension objects defining the set"""

    @model_validator(mode="after")
    def no_repeated_dimensions(self):
        """Check that all dimensions have unique letters."""
        letters = self.letters
        if len(letters) != len(set(letters)):
            raise ValueError("Dimensions must have unique letters in DimensionSet.")
        return self

    @property
    def _dict(self) -> Dict[str, Dimension]:
        """Contains mappings.

        letter --> dim object and name --> dim object
        """
        return {dim.name: dim for dim in self.dim_list} | {dim.letter: dim for dim in self.dim_list}

    def __getitem__(self, key) -> Dimension:
        """Get a dimension by its name, letter or index with the [] operator.

        Args:
            key (str, int, tuple): The name, letter or index of the dimension to get; If a tuple is passed, a subset of the dimensions is returned as a new DimensionSet.

        Example:
        >>> dimensions['Region']
        >>> dimensions['r']
        >>> dimensions[0]
        >>> dimensions['Region', 'Time']
        """
        if isinstance(key, tuple):
            return self.get_subset(key)
        if isinstance(key, str):
            return self._dict[key]
        elif isinstance(key, int):
            return self.dim_list[key]
        else:
            raise TypeError("Key must be string or int")

    def __iter__(self) -> Iterator[Dimension]:
        return iter(self.dim_list)

    def __contains__(self, key: str) -> bool:
        return key in self._dict

    def size(self, key: str):
        """get the number of items in a dimension

        Args:
            key (str): the name or letter of the dimension to get the size of
        """
        return self._dict[key].len

    def shape(self, keys: tuple = None):
        """get the shape of the array that would be created with the dimensions in the set

        Args:
            keys (tuple, optional): the names or letters of the dimensions to get the shape of. If None, all dimensions are used. Defaults to None.
        """
        keys = keys if keys else self.letters
        return tuple(self.size(key) for key in keys)

    @property
    def ndim(self):
        """the number of dimensions in the set"""
        return len(self.dim_list)

    def get_subset(self, dims: tuple = None) -> "DimensionSet":
        """Selects :py:class:`Dimension` objects from the object attribute dim_list,
        according to the dims passed, which can be either letters or names.
        Returns a copy if dims are not given.
        """
        subset = copy(self)
        if dims is not None:
            subset.dim_list = [self._dict[dim_key] for dim_key in dims]
        return subset

    def expand_by(self, added_dims: list[Dimension]) -> "DimensionSet":
        """Expands the DimensionSet by adding new dimensions to it."""
        if not all([dim.letter not in self.letters for dim in added_dims]):
            raise ValueError(
                "DimensionSet already contains one or more of the dimensions to be added."
            )
        return DimensionSet(dim_list=self.dim_list + added_dims)

    def drop(self, key: str, inplace: bool = False) -> Optional["DimensionSet"]:
        """Remove a dimension from the set.

        Args:
            key (str): The name, index or letter of the dimension to drop
            inplace (bool, optional): If True, the operation is performed in place, otherwise a new DimensionSet is returned. Defaults to False.

        Returns:
            None if inplace=True, otherwise a new DimensionSet with the dimension removed
        """
        dim_to_drop = self._dict[key]
        if inplace:
            self.dim_list.remove(dim_to_drop)
            return
        else:
            dimensions = copy(self.dim_list)
            dimensions.remove(dim_to_drop)
            return DimensionSet(dim_list=dimensions)

    def replace(self, key: str, new_dim: Dimension, inplace: bool = False):
        """Replace a dimension in the set with a new one.

        Args:
            key (str): The name, index or letter of the dimension to replace
            new_dim (Dimension): The new dimension to replace the old one
            inplace (bool, optional): If True, the operation is performed in place, otherwise a new DimensionSet is returned. Defaults to False.

        Returns:
            None if inplace=True, otherwise a new DimensionSet with the dimension replaced
        """
        if new_dim.letter in self.letters:
            raise ValueError(
                "New dimension can't have same letter as any of those already in DimensionSet, "
                "as that would create ambiguity"
            )
        if inplace:
            self.dim_list[self.index(key)] = new_dim
            return
        else:
            dim_list = copy(self.dim_list)
            dim_list[self.index(key)] = new_dim
            return DimensionSet(dim_list=dim_list)

    def intersect_with(self, other: "DimensionSet") -> "DimensionSet":
        """Get the intersection of two DimensionSets.

        Args:
            other (DimensionSet): The other DimensionSet to intersect with

        Returns:
            DimensionSet: The intersection of the two DimensionSets
        """
        intersection_letters = [dim.letter for dim in self.dim_list if dim.letter in other.letters]
        return self.get_subset(intersection_letters)

    def union_with(self, other: "DimensionSet") -> "DimensionSet":
        """Get the union of two DimensionSets.

        Args:
            other (DimensionSet): The other DimensionSet to unite with

        Returns:
            DimensionSet: The union of the two DimensionSets
        """
        added_dims = [dim for dim in other.dim_list if dim.letter not in self.letters]
        return self.expand_by(added_dims)

    def difference_with(self, other: "DimensionSet") -> "DimensionSet":
        """Get the set difference of two DimensionSets.

        Args:
            other (DimensionSet): The other DimensionSet to compare with

        Returns:
            DimensionSet: The difference of the two DimensionSets
        """
        difference_letters = [
            dim.letter for dim in self.dim_list if dim.letter not in other.letters
        ]
        return self.get_subset(difference_letters)

    @property
    def names(self):
        """A tuple of the names of the dimensions in the set."""
        return tuple([dim.name for dim in self.dim_list])

    @property
    def letters(self):
        """A tuple of the letters of the dimensions in the set."""
        return tuple([dim.letter for dim in self.dim_list])

    @property
    def string(self):
        """The letters of the dimensions in the set concatenated to a single string."""
        return "".join(self.letters)

    def index(self, key):
        """Return the index of a dimension in the set."""
        return [d.letter for d in self.dim_list].index(key)



from typing import Callable

from .processes import Process
from .flodym_arrays import Flow
from .dimensions import DimensionSet
from .mfa_definition import FlowDefinition
from .flow_naming import process_names_with_arrow


def make_empty_flows(
    processes: dict[str, Process],
    flow_definitions: list[FlowDefinition],
    dims: DimensionSet,
    naming: Callable[[Process, Process], str] = process_names_with_arrow,
) -> dict[str, Flow]:
    """Initialize all defined flows with zero values.

    Args:
        processes: Dictionary of processes, with process names as keys.
        flow_definitions: List of flow definitions.
        dims: DimensionSet object containing all dimensions.
        naming: Function to generate names for flows. Default is `process_names_with_arrow`.

    Returns:
        Dictionary of flows, with flow names as keys.
    """
    flows = {}
    for flow_definition in flow_definitions:
        try:
            from_process = processes[flow_definition.from_process_name]
            to_process = processes[flow_definition.to_process_name]
        except KeyError:
            raise KeyError(f"Missing process required by flow definition {flow_definition}.")
        if flow_definition.name_override is not None:
            name = flow_definition.name_override
        else:
            name = naming(from_process, to_process)
        dim_subset = dims.get_subset(flow_definition.dim_letters)
        flow = Flow(from_process=from_process, to_process=to_process, name=name, dims=dim_subset)
        flows[flow.name] = flow
    return flows



import re
import unicodedata
from pydantic import BaseModel as PydanticBaseModel
from typing import Optional


def to_valid_file_name(value: str) -> str:
    """
    Taken from https://github.com/django/django/blob/master/django/utils/text.py
    Convert to ASCII. Convert spaces or repeated dashes to single dashes.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]", "_", value).strip("-_")


class CustomNameDisplayer(PydanticBaseModel):
    """
    Parameters:
        display_names: Dictionary for string replacement in figures. All strings not in this dictionary will be displayed as is.
    """

    display_names: Optional[dict] = {}
    """Dictionary for string replacement in figures. Keys are strings to be replaced (like process names, etc.), values are strings to display instead.
    All strings not in this dictionary will be displayed as is.
    """

    def display_name(self, name):
        return self.display_names[name] if name in self.display_names else name



"""Home to helper functions for working with `FlodymArray`s."""

from .flodym_arrays import FlodymArray
from .dimensions import Dimension


def flodym_array_stack(flodym_arrays: list[FlodymArray], dimension: Dimension) -> FlodymArray:
    """Stack a list of FlodymArray objects using a new dimension.
    Like numpy.stack with axis=-1, but for `FlodymArray`s.
    Method can be applied to `FlodymArray`s, `StockArray`s, `Parameter`s and `Flow`s.
    """
    flodym_array0 = flodym_arrays[0]
    extended_dimensions = flodym_array0.dims.expand_by([dimension])
    extended = FlodymArray(dims=extended_dimensions)
    for item, flodym_array in zip(dimension.items, flodym_arrays):
        extended[{dimension.letter: item}] = flodym_array
    return extended



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal, assert_array_equal
from pydantic_core import ValidationError
import pytest

from flodym import FlodymArray, DimensionSet, Dimension


places = Dimension(name="place", letter="p", items=["Earth", "Sun", "Moon", "Venus"])
local_places = Dimension(name="local place", letter="l", items=["Earth"])
time = Dimension(name="time", letter="t", items=[1990, 2000, 2010])
historic_time = Dimension(name="historic time", letter="h", items=[1990, 2000])
animals = Dimension(name="animal", letter="a", items=["cat", "mouse"])

base_dim_list = [places, time]

dims = DimensionSet(dim_list=base_dim_list)
values = np.random.rand(4, 3)
numbers = FlodymArray(name="two", dims=dims, values=values)

dims_incl_animals = DimensionSet(dim_list=base_dim_list + [animals])
animal_values = np.random.rand(4, 3, 2)
space_animals = FlodymArray(name="space_animals", dims=dims_incl_animals, values=animal_values)


def test_flodym_array_validations():
    dims = DimensionSet(dim_list=[local_places, time])

    # example with values with the correct shape
    FlodymArray(name="numbers", dims=dims, values=np.array([[1, 2, 3]]))

    # example with dimensions reversed
    with pytest.raises(ValidationError):
        FlodymArray(name="numbers", dims=dims, values=np.array([[1], [2], [3]]))

    # example with too many values
    with pytest.raises(ValidationError):
        FlodymArray(name="numbers", dims=dims, values=np.array([[1, 2, 3, 4]]))

    # example with no values passed -> filled with zeros
    zero_values = FlodymArray(name="numbers", dims=dims)
    assert zero_values.values.shape == (1, 3)
    assert np.all([zero_values.values == 0])


def test_cast_to():
    # example of duplicating values along new axis (e.g. same number of cats and mice)
    casted_flodym_array = numbers.cast_to(target_dims=dims_incl_animals)
    assert casted_flodym_array.dims == dims_incl_animals
    assert casted_flodym_array.values.shape == (4, 3, 2)
    assert_almost_equal(np.sum(casted_flodym_array.values), 2 * np.sum(values))

    # example with differently ordered dimensions
    target_dims = DimensionSet(dim_list=[animals] + base_dim_list[::-1])
    casted_flodym_array = numbers.cast_to(target_dims=target_dims)
    assert casted_flodym_array.values.shape == (2, 3, 4)


def test_sum_to():
    # sum over one dimension
    summed_flodym_array = space_animals.sum_to(result_dims=("p", "t"))
    assert summed_flodym_array.dims == DimensionSet(dim_list=base_dim_list)
    assert_array_almost_equal(summed_flodym_array.values, np.sum(animal_values, axis=2))

    # sum over two dimensions
    summed_flodym_array = space_animals.sum_to(result_dims=("t"))
    assert_array_almost_equal(
        summed_flodym_array.values, np.sum(np.sum(animal_values, axis=2), axis=0)
    )

    # example attempt to get a resulting dimension that does not exist
    with pytest.raises(KeyError):
        space_animals.sum_to(result_dims=("s"))

    # example where dimensions to sum over are specified rather than the remaining dimensions
    summed_over = space_animals.sum_over(sum_over_dims=("p", "a"))
    assert_array_almost_equal(summed_over.values, summed_flodym_array.values)

    # example sum over dimension that doesn't exist
    with pytest.raises(KeyError):
        space_animals.sum_over(sum_over_dims=("s"))


def test_get_shares_over():
    # example of getting shares over one dimension
    shares = space_animals.get_shares_over(dim_letters=("p"))
    assert shares.dims == space_animals.dims
    wanted_values = np.einsum("pta,ta->pta", animal_values, 1 / np.sum(animal_values, axis=0))
    assert_array_almost_equal(shares.values, wanted_values)

    # example of getting shares over two dimensions
    shares = space_animals.get_shares_over(dim_letters=("p", "a"))
    wanted_values = np.einsum("pta,t->pta", animal_values, 1 / np.sum(animal_values, axis=(0, 2)))
    assert_array_almost_equal(shares.values, wanted_values)

    # example of getting shares over all dimensions
    shares = space_animals.get_shares_over(dim_letters=("p", "t", "a"))
    assert_array_almost_equal(shares.values, animal_values / np.sum(animal_values))

    # example of getting shares over a dimension that doesn't exist
    with pytest.raises(AssertionError):
        space_animals.get_shares_over(dim_letters=("s",))


def test_maths():
    # test minimum
    minimum = space_animals.minimum(numbers)
    assert minimum.dims == dims
    assert_array_almost_equal(minimum.values, np.minimum(values, animal_values.sum(axis=2)))

    # test maximum
    maximum = space_animals.maximum(numbers)
    assert maximum.dims == dims
    assert_array_almost_equal(maximum.values, np.maximum(values, animal_values.sum(axis=2)))

    # test sum
    summed = space_animals + numbers
    assert summed.dims == dims
    assert_array_almost_equal(summed.values, animal_values.sum(axis=2) + values)

    # test minus
    subtracted = space_animals - numbers
    assert subtracted.dims == dims
    assert_array_almost_equal(subtracted.values, animal_values.sum(axis=2) - values)
    subtracted_flipped = numbers - space_animals
    assert subtracted_flipped.dims == dims
    assert_array_almost_equal(subtracted_flipped.values, values - animal_values.sum(axis=2))

    # test multiply
    multiplied = numbers * space_animals
    assert multiplied.dims == dims_incl_animals  # different from behaviour of above methods
    assert_array_almost_equal(multiplied.values[:, :, 0], values * animal_values[:, :, 0])
    assert_array_almost_equal(multiplied.values[:, :, 1], values * animal_values[:, :, 1])

    # test divide
    divided = space_animals / numbers
    assert divided.dims == dims_incl_animals
    assert_array_almost_equal(divided.values[:, :, 0], animal_values[:, :, 0] / values)
    assert_array_almost_equal(divided.values[:, :, 1], animal_values[:, :, 1] / values)
    divided_flipped = numbers / space_animals
    assert divided_flipped.dims == dims_incl_animals
    assert_array_almost_equal(divided_flipped.values[:, :, 0], values / (animal_values[:, :, 0]))
    assert_array_almost_equal(divided_flipped.values[:, :, 1], values / (animal_values[:, :, 1]))


def test_get_item():
    cats_on_the_moon = space_animals["Moon"]["cat"]
    assert isinstance(cats_on_the_moon, FlodymArray)
    assert_array_almost_equal(cats_on_the_moon.values, space_animals.values[2, :, 0])
    # note that this does not work for the time dimension (not strings)
    # and also assumes that no item appears in more than one dimension


def test_sub_array_handler():
    space_cat = space_animals["cat"]  # space cat from str
    another_space_cat = space_animals[{"a": "cat"}]  # space cat from dict
    assert_array_equal(space_cat.values, another_space_cat.values)

    space_1990 = space_animals[{"t": 1990}]  # space animals in 1990
    assert space_1990.values.shape == (4, 2)
    assert space_1990.dims.letters == ("p", "a")

    with pytest.raises(ValueError):
        space_animals[{"a": "dog"}]  # there isn't a dog in space_animals


def test_dimension_subsets():
    historic_dims_incl_animals = DimensionSet(dim_list=[places, historic_time, animals])
    historic_space_animals = FlodymArray(dims=historic_dims_incl_animals)
    historic_space_animals[...] = space_animals[{"t": historic_time}]

    assert np.min(historic_space_animals.values) > 0.0

    space_animals_copy = FlodymArray(dims=dims_incl_animals)
    space_animals_copy[{"t": historic_time}] = 1.0 * historic_space_animals
    space_animals_copy[{"t": 2010}] = 1.0 * space_animals[{"t": 2010}]

    assert_array_equal(space_animals.values, space_animals_copy.values)

    wrong_historic_time = (Dimension(name="historic time", letter="t", items=[1990, 2000]),)
    with pytest.raises(ValueError):
        space_animals[{"t": wrong_historic_time}]  # same letter in original and subset

    another_wrong_historic_time = (Dimension(name="historic time", letter="p", items=[1990, 2000]),)
    with pytest.raises(ValueError):
        space_animals[
            {"t": another_wrong_historic_time}
        ]  # same letter in other dim of original and subset

    with pytest.raises(ValueError):
        historic_space_animals[{"h": time}]  # index is not a subset



from pydantic_core import ValidationError
import pytest

from flodym import DimensionSet


def test_validate_dimension_set():
    # example valid DimensionSet
    dimensions = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    DimensionSet(dim_list=dimensions)

    # example with repeated dimension letters in DimensionSet
    dimensions.append({"name": "another_time", "letter": "t", "items": [2020, 2030]})
    with pytest.raises(ValidationError) as error_msg:
        DimensionSet(dim_list=dimensions)
    assert "letter" in str(error_msg.value)


def test_get_subset():
    subset_dimensions = [
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
    material_dimension = {"name": "material", "letter": "m", "items": ["material_0", "material_1"]}

    parent_dimensions = subset_dimensions + [material_dimension]
    dimension_set = DimensionSet(dim_list=parent_dimensions)

    # example of subsetting the dimension set using dimension letters
    subset_from_letters = dimension_set.get_subset(dims=("t", "p"))
    assert subset_from_letters == DimensionSet(dim_list=subset_dimensions)

    # example of subsetting the dimension set using dimension names
    subset_from_names = dimension_set.get_subset(dims=("time", "place"))
    assert subset_from_names == subset_from_letters

    # example where the requested subset dimension doesn't exist
    with pytest.raises(KeyError):
        dimension_set.get_subset(dims=("s", "p"))
